/*
 * Copyright (c) 1994 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Dlg/dlgcov.h>
#include <Dlg/dlgdb.h>
#include <Dlg/dlgfclass.h>
#include <Dlg/dlgfeature.h>
#include <Dlg/dlglib.h>

#include <Map/mapprim.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <Unidraw/catalog.h>
#include <Unidraw/unidraw.h>

#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/Graphic/pspaint.h>

#include <OS/math.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>

/*****************************************************************************/

DlgFeatureClass::DlgFeatureClass (DlgCoverage* library, const char* name) 
: MapFeatureClass(library, name) {
    proj = nil;
    _extent = false;
}

DlgFeatureClass::~DlgFeatureClass ()
{
}

DlgFeature* DlgFeatureClass::feature(unsigned fclsnum) const {
    return (DlgFeature*)MapFeatureClass::feature(fclsnum);
}

const char* DlgFeatureClass::description() const {
    sprintf(((DlgFeatureClass*)this)->dlgdesc, "%s feature class", name());
    return dlgdesc;
}

PJ* DlgFeatureClass::projection() {
    if (!proj)
	projection(make_projection());
    return proj;
}

void DlgFeatureClass::projection(PJ* p) { proj = p; }

PJ* DlgFeatureClass::make_projection() {
  if (!proj) {
    if (((DlgDatabase*)(this->coverage()->library()->database()))->Is100K()) {
	return nil;
    }
    else {
    char shcmd[strlen("gunzip -c ") + 1 + strlen(coverage()->path())];
    if (strstr(coverage()->path(), ".Z"))
	strcpy( shcmd, "zcat ");
    else if ( strstr(coverage()->path(), ".gz"))
	strcpy( shcmd, "gunzip -c ");
    else
	strcpy( shcmd, "cat ");
    strcat( shcmd, coverage()->path());

    FILE* fptr  = popen( shcmd, "r");
    if (!fptr) {
	return nil;
    }
    char buffer[145];
    buffer[144] = '\0';

    /* "A" records */
    char first_std_lat[40];
    char second_std_lat[40];
    char central_meridian_lon[40];
    char proj_origin_lat[40];
    for (int recnum=0; recnum<6; recnum++) {
	fread (buffer, 1, 144, fptr);
	char degbuf[3], minbuf[3], secbuf[11];
	if (recnum == 1) { // record A.2
	    // latitude of first standard parallel
	    memcpy(degbuf, buffer+18+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(minbuf, buffer+18+24+24 +8, 2); minbuf[2] = '\0';
	    memcpy(secbuf, buffer+18+24+24 +11, 2); secbuf[2] = '.';
	    memcpy(secbuf+3, buffer+18+24+24 +13, 7); secbuf[10] = '\0';
	    strcpy(first_std_lat, "+lat_1=");
	    strcat(first_std_lat, degbuf);
	    strcat(first_std_lat, "d");
	    strcat(first_std_lat, minbuf);
	    strcat(first_std_lat, "'");
	    strcat(first_std_lat, secbuf);
	    strcat(first_std_lat, "\"");
	    if (buffer[18+24+24 +2] == '-')
		strcat(first_std_lat, "S");
	    else
		strcat(first_std_lat, "N");

	    // latitude of second standard parallel
	    memcpy(degbuf, buffer+18+24+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(degbuf, buffer+18+24+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(minbuf, buffer+18+24+24+24 +8, 2); minbuf[2] = '\0';
	    memcpy(secbuf, buffer+18+24+24+24 +11, 2); secbuf[2] = '.';
	    memcpy(secbuf+3, buffer+18+24+24+24 +13, 7); secbuf[10] = '\0';
	    strcpy(second_std_lat, "+lat_2=");
	    strcat(second_std_lat, degbuf);
	    strcat(second_std_lat, "d");
	    strcat(second_std_lat, minbuf);
	    strcat(second_std_lat, "'");
	    strcat(second_std_lat, secbuf);
	    strcat(second_std_lat, "\"");
	    if (buffer[18+24+24+24 +2] == '-')
		strcat(second_std_lat, "S");
	    else
		strcat(second_std_lat, "N");

	    // longitude of central meridian
	    memcpy(degbuf, buffer+18+24+24+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(degbuf, buffer+18+24+24+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(minbuf, buffer+18+24+24+24+24 +8, 2); minbuf[2] = '\0';
	    memcpy(secbuf, buffer+18+24+24+24+24 +11, 2); secbuf[2] = '.';
	    memcpy(secbuf+3, buffer+18+24+24+24+24 +13, 7); secbuf[10] = '\0';
	    strcpy(central_meridian_lon, "+lon_0=");
	    strcat(central_meridian_lon, degbuf);
	    strcat(central_meridian_lon, "d");
	    strcat(central_meridian_lon, minbuf);
	    strcat(central_meridian_lon, "'");
	    strcat(central_meridian_lon, secbuf);
	    strcat(central_meridian_lon, "\"");
	    if (buffer[18+24+24+24+24 +2] == '-')
		strcat(central_meridian_lon, "W");
	    else
		strcat(central_meridian_lon, "E");
	}
	else if (recnum == 2) { // record A.3
	    // latitude of projection origin
	    memcpy(degbuf, buffer +5, 2); degbuf[2] = '\0';
	    memcpy(minbuf, buffer +8, 2); minbuf[2] = '\0';
	    memcpy(secbuf, buffer +11, 2); secbuf[2] = '.';
	    memcpy(secbuf+3, buffer +13, 7); secbuf[10] = '\0';
	    strcpy(proj_origin_lat, "+lat_0=");
	    strcat(proj_origin_lat, degbuf);
	    strcat(proj_origin_lat, "d");
	    strcat(proj_origin_lat, minbuf);
	    strcat(proj_origin_lat, "'");
	    strcat(proj_origin_lat, secbuf);
	    strcat(proj_origin_lat, "\"");
	    if (buffer[2] == '-')
		strcat(proj_origin_lat, "S");
	    else
		strcat(proj_origin_lat, "N");
	}
    }
    char* params[7];
    params[0] = "proj=aea";
    params[1] = "ellps=clrk66";
    params[2] = central_meridian_lon;
    params[3] = proj_origin_lat;
    params[4] = first_std_lat;
    params[5] = second_std_lat;
    params[6] = "no_defs";

    do { fread (buffer, 1, 144, fptr); } while (!feof(fptr));
    pclose(fptr);
    proj = pj_init(7, params);
}
  }
  return proj;
}

double power_of_ten(double orig, int exponent) {
    double result = orig;
    for (int i = 0; i < abs(exponent); i++)
	result *= exponent > 0 ? 10.0 : 1.0/10.0;
    return result;
}

MapFeature** DlgFeatureClass::make_features(int &nfeats) { 
    if (((DlgDatabase*)(this->coverage()->library()->database()))->Is100K()) {
	int fnum;
	DlgFeature** features =nil;
	FILE* fptr = fopen(path(), "r");
	if (fptr) {
	  char buffer[83];
	  buffer[82] = '\0';

	  for (int recnum=0; recnum<4; recnum++) {
	    fgets (buffer, 82, fptr);
	  }
	  int num_ctrlpts = atoi(buffer+54);
	  int num_categories = atoi(buffer+60);

	  for (int recnum=0; recnum<6+num_ctrlpts+num_categories; recnum++)
	    fgets (buffer, 82, fptr);
	  int max_nodes = atoi(buffer+24);
	  int actual_nodes = atoi(buffer+30);
	  boolean node_area_links = buffer[37] == '1';
	  boolean node_line_links = buffer[38] == '1';
	  int max_faces = atoi(buffer+40);
	  int actual_faces = atoi(buffer+46);
	  boolean area_node_links = buffer[53] == '1';
	  boolean area_line_links = buffer[54] == '1';
	  boolean area_coords = buffer[55] == '1';
	  int max_edges = atoi(buffer+56);
	  int actual_edges = atoi(buffer+62);
	  boolean line_coords = buffer[71] == '1';

	  /* allocate arrays for primitives */
	  MapNode** nodes = new MapNode*[actual_nodes];
	  MapFace** faces = new MapFace*[actual_faces];
	  MapEdge** edges = new MapEdge*[actual_edges];
	  int nnodes = actual_nodes;
	  int nfaces = actual_faces;
	  int nedges = actual_edges;
	    
	  int nodenum = 0;
	  int facenum = 0;
	  int edgenum = 0;
	    
	  int node_index[max_nodes+1];
	  int face_index[max_faces+1];
	  int edge_index[max_edges+1];

	  int i;
	  for (i=0; i<max_nodes; i++) 
	    node_index[i] = -1;
	  for (i=0; i<max_faces; i++) 
	    face_index[i] = -1;
	  for (i=0; i<max_edges; i++) 
	    edge_index[i] = -1;
	    
	  int nprims = nnodes + nfaces + nedges;
	  fnum = 0;
	  features = new DlgFeature*[nprims];
	  for (i = 0; i < nprims; i++)
	    features[i] = nil;

	  // nodes
	    for (i = 0; i < actual_nodes; i++) {
		while (buffer[0] != 'N')
		    fgets(buffer, 82, fptr);
		int id = atoi(buffer + 1);
		float nx, ny;
		nx = atof(buffer + 6);
		ny = atof(buffer + 18);
		//int numlines = atoi(buffer + 36);
		int numlines = atoi(buffer + 30);
		int nattrpairs = atoi(buffer + 48);
		if (node_area_links) { // never there, they say
		    fgets(buffer, 82, fptr);
		}
		int connections[12];
		int nconn = 0;
		if (node_line_links) {
		    fgets(buffer, 82, fptr);
		    for (int j = 0; j < numlines; j++) {
		      char ibuf[7];
		      ibuf[6] = '\0';
		      memcpy(ibuf, buffer+(j*6), 6);
		      int lineid = atoi(ibuf);
		      connections[nconn++] = lineid;
		    }
		}
		int* majorcodes = nil;
		int* minorcodes = nil;
		boolean is_feature = false;
		if (nattrpairs > 0) {
		    majorcodes = new int[nattrpairs];
		    minorcodes = new int[nattrpairs];
		    fgets(buffer, 82, fptr);
		    for (int j = 0; j < nattrpairs; j++) {
			majorcodes[j] = atoi(buffer + (j * 6));
			minorcodes[j] = atoi(buffer + ((j+1) * 6));
		    }
		    is_feature = true;
		}
		if (is_feature) {
		    features[fnum] = new DlgFeature((DlgFeatureClass*)this, id,
						    MapFeature::PointType);
		    if (majorcodes) {
			// just first one for now
			features[fnum]->set_attributes(majorcodes[0], minorcodes[0]);
		    }
		}
		else {
		  features[fnum] = new DlgFeature((DlgFeatureClass*)this, id,
						  MapFeature::JunctionType);
		  AttributeValueList* connlist = new AttributeValueList();
		  for (int j = 0; j < nconn; j++) {
		    AttributeValue* val = new AttributeValue(connections[j],
							     AttributeValue::IntType);
		    connlist->Append(val);
		  }
		  AttributeValue* connattr = new AttributeValue(connlist);
		  features[fnum]->attrlist()->Append(new Attribute("connections",
								   connattr));
		}
		nodes[nodenum] = new MapNode(/*is_feature ?*/ features[fnum] /*: nil*/);
		nodes[nodenum]->insert_points(1, &nx, &ny);
		node_index[id] = nodenum;
		//if (is_feature) {
		    features[fnum]->node_primitive(nodes[nodenum]);
		    fnum++;
		    //}
		nodenum++;
	    }

	    // areas
	    for (i = 0; i < actual_faces; i++) {
		while (buffer[0] != 'A')
		    fgets(buffer, 82, fptr);
		int id = atoi(buffer + 1);
		float nx, ny;
		nx = atof(buffer + 6);
		ny = atof(buffer + 18);
		int numlines = atoi(buffer + 36);
		int nattrpairs = atoi(buffer + 48);
		if (area_node_links) { // never there, they say
		    fgets(buffer, 82, fptr);
		}
		if (area_line_links) {
		    int nrec = (int)ceil(numlines/12.);
		    for (int j = 0; j < nrec; j++) {
			fgets(buffer, 82, fptr);
		    }
		}
		int* majorcodes = nil;
		int* minorcodes = nil;
		boolean is_feature = false;
		if (nattrpairs > 0) {
		    majorcodes = new int[nattrpairs];
		    minorcodes = new int[nattrpairs];
		    fgets(buffer, 82, fptr);
		    for (int j = 0; j < nattrpairs; j++) {
			majorcodes[j] = atoi(buffer + (j * 6));
			minorcodes[j] = atoi(buffer + ((j+1) * 6));
		    }
		    if (majorcodes[0] == 0 && minorcodes[0] == 0)
			is_feature = false;
		    else
			is_feature = true;
		}
		if (is_feature) {
		    features[fnum] = new DlgFeature((DlgFeatureClass*)this, id,
						    MapFeature::AreaType);
		    if (majorcodes) {
			// just first one for now
			features[fnum]->set_attributes(majorcodes[0], minorcodes[0]);
		    }
		}
		faces[facenum] = new MapFace(is_feature ? features[fnum] : nil);
		face_index[id] = facenum;
		if (is_feature) {
		    features[fnum]->face_primitive(faces[facenum]);
		    fnum++;
		}
		facenum++;
	    }

	    // lines
	    for (i = 0; i < actual_edges; i++) {
		while (buffer[0] != 'L')
		    fgets(buffer, 82, fptr);
		int id = atoi(buffer + 1);
		int start_node_id = atoi(buffer+6);
		int end_node_id = atoi(buffer+12);
		int left_face_id = atoi(buffer + 18);
		int right_face_id = atoi(buffer + 24);
		int npts = atoi(buffer + 42);
		int nattrpairs = atoi(buffer + 48);

		/* read in all the edge coordinate pairs */
		float *x = new float[npts];
		float *y = new float[npts];
		int nrec = (int)ceil(npts/3.);
		int ptnum = 0;
		for (int rec=0; rec<nrec; rec++) {
		    fgets (buffer, 82, fptr);
		    int count = 0;
		    while (count < 3 && ptnum<npts) {
			x[ptnum] = (float)atoi(buffer+count*24);
			y[ptnum] = (float)atoi(buffer+count*24+12);
			count++;
			ptnum++;
		    }
		}

		/* read in all the attribute code pairs */
		int* majorcodes = nil;
		int* minorcodes = nil;
		boolean is_feature = false;
		if (nattrpairs > 0) {
		    majorcodes = new int[nattrpairs];
		    minorcodes = new int[nattrpairs];
		    fgets(buffer, 82, fptr);
		    for (int j = 0; j < nattrpairs; j++) {
			majorcodes[j] = atoi(buffer + (j * 6));
			minorcodes[j] = atoi(buffer + ((j+1) * 6));
		    }
		    is_feature = true;
		}
		if (is_feature) {
		    features[fnum] = new DlgFeature((DlgFeatureClass*)this, id,
						    MapFeature::LineType);
		    if (majorcodes) {
			// just first one for now
			features[fnum]->set_attributes(majorcodes[0], minorcodes[0]);
		    }
		}
		MapEdge* edge = new MapEdge;
		edges[edgenum] = edge;
		edges[edgenum]->insert_pointers(npts, x, y, nil, true);
		if (is_feature) {
		    features[fnum]->edge_primitive(edges[edgenum]);
		    fnum++;
		}
		edgenum++;

		MapNode* start_node;
		if (node_index[start_node_id]<0) {
		    node_index[start_node_id] = nodenum;
		    start_node = new MapNode();
		    nodes[nodenum++] = start_node;
		} else 
		    start_node = nodes[node_index[start_node_id]];
		MapNode* end_node;
		if (node_index[end_node_id]<0) {
		    node_index[end_node_id] = nodenum;
		    end_node = new MapNode();
		    nodes[nodenum++] = end_node;
		} else 
		    end_node = nodes[node_index[end_node_id]];
		edge->attach_nodes(start_node, end_node);

		MapFace* left_face;
		if (face_index[left_face_id]<0) {
		    face_index[left_face_id] = facenum;
		    left_face = new MapFace();
		    faces[facenum++] = left_face;
		} else 
		    left_face = faces[face_index[left_face_id]];
		MapFace* right_face;
		if (face_index[right_face_id]<0) {
		    face_index[right_face_id] = facenum;
		    right_face = new MapFace();
		    faces[facenum++] = right_face;
		} else 
		    right_face = faces[face_index[right_face_id]];
		edge->attach_faces(left_face, right_face);
	    }
	    fclose(fptr);
	}
	nfeats = fnum;
	return (MapFeature**)features;
    }
    else {
    char shcmd[strlen("gunzip -c ") + 1 + strlen(coverage()->path())];
    if (strstr(coverage()->path(), ".Z"))
	strcpy( shcmd, "zcat ");
    else if ( strstr(coverage()->path(), ".gz"))
	strcpy( shcmd, "gunzip -c ");
    else
	strcpy( shcmd, "cat ");
    strcat( shcmd, coverage()->path());

    FILE* fptr  = popen( shcmd, "r");
    if (!fptr) {
	nfeats  = 0;
	return nil;
    }
    char buffer[145];
    buffer[144] = '\0'; //make sure null terminated string for atoi below

    /* "A" records */
    char first_std_lat[40];
    char second_std_lat[40];
    char central_meridian_lon[40];
    char proj_origin_lat[40];
    for (int recnum=0; recnum<6; recnum++) {
	fread (buffer, 1, 144, fptr);
	char degbuf[3], minbuf[3], secbuf[11];
	if (recnum == 1) { // record A.2
	    // latitude of first standard parallel
	    memcpy(degbuf, buffer+18+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(minbuf, buffer+18+24+24 +8, 2); minbuf[2] = '\0';
	    memcpy(secbuf, buffer+18+24+24 +11, 2); secbuf[2] = '.';
	    memcpy(secbuf+3, buffer+18+24+24 +13, 7); secbuf[10] = '\0';
	    strcpy(first_std_lat, "+lat_1=");
	    strcat(first_std_lat, degbuf);
	    strcat(first_std_lat, "d");
	    strcat(first_std_lat, minbuf);
	    strcat(first_std_lat, "'");
	    strcat(first_std_lat, secbuf);
	    strcat(first_std_lat, "\"");
	    if (buffer[18+24+24 +2] == '-')
		strcat(first_std_lat, "S");
	    else
		strcat(first_std_lat, "N");

	    // latitude of second standard parallel
	    memcpy(degbuf, buffer+18+24+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(degbuf, buffer+18+24+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(minbuf, buffer+18+24+24+24 +8, 2); minbuf[2] = '\0';
	    memcpy(secbuf, buffer+18+24+24+24 +11, 2); secbuf[2] = '.';
	    memcpy(secbuf+3, buffer+18+24+24+24 +13, 7); secbuf[10] = '\0';
	    strcpy(second_std_lat, "+lat_2=");
	    strcat(second_std_lat, degbuf);
	    strcat(second_std_lat, "d");
	    strcat(second_std_lat, minbuf);
	    strcat(second_std_lat, "'");
	    strcat(second_std_lat, secbuf);
	    strcat(second_std_lat, "\"");
	    if (buffer[18+24+24+24 +2] == '-')
		strcat(second_std_lat, "S");
	    else
		strcat(second_std_lat, "N");

	    // longitude of central meridian
	    memcpy(degbuf, buffer+18+24+24+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(degbuf, buffer+18+24+24+24+24 +5, 2); degbuf[2] = '\0';
	    memcpy(minbuf, buffer+18+24+24+24+24 +8, 2); minbuf[2] = '\0';
	    memcpy(secbuf, buffer+18+24+24+24+24 +11, 2); secbuf[2] = '.';
	    memcpy(secbuf+3, buffer+18+24+24+24+24 +13, 7); secbuf[10] = '\0';
	    strcpy(central_meridian_lon, "+lon_0=");
	    strcat(central_meridian_lon, degbuf);
	    strcat(central_meridian_lon, "d");
	    strcat(central_meridian_lon, minbuf);
	    strcat(central_meridian_lon, "'");
	    strcat(central_meridian_lon, secbuf);
	    strcat(central_meridian_lon, "\"");
	    if (buffer[18+24+24+24+24 +2] == '-')
		strcat(central_meridian_lon, "W");
	    else
		strcat(central_meridian_lon, "E");
	}
	else if (recnum == 2) { // record A.3
	    // latitude of projection origin
	    memcpy(degbuf, buffer +5, 2); degbuf[2] = '\0';
	    memcpy(minbuf, buffer +8, 2); minbuf[2] = '\0';
	    memcpy(secbuf, buffer +11, 2); secbuf[2] = '.';
	    memcpy(secbuf+3, buffer +13, 7); secbuf[10] = '\0';
	    strcpy(proj_origin_lat, "+lat_0=");
	    strcat(proj_origin_lat, degbuf);
	    strcat(proj_origin_lat, "d");
	    strcat(proj_origin_lat, minbuf);
	    strcat(proj_origin_lat, "'");
	    strcat(proj_origin_lat, secbuf);
	    strcat(proj_origin_lat, "\"");
	    if (buffer[2] == '-')
		strcat(proj_origin_lat, "S");
	    else
		strcat(proj_origin_lat, "N");
	}
    }
    char* params[7];
    params[0] = "proj=aea";
    params[1] = "ellps=clrk66";
    params[2] = central_meridian_lon;
    params[3] = proj_origin_lat;
    params[4] = first_std_lat;
    params[5] = second_std_lat;
    params[6] = "no_defs";
    proj = pj_init(7, params);

    /* "B" records */
    double A1, A2, A3, A4;
    int D1, D2, D3, D4;
    for (int recnum=0; recnum<2; recnum++) {
	fread (buffer, 1, 144, fptr);
	if (recnum == 0) {
	    A1 = atof(buffer);
	    D1 = atoi(buffer+21);
	    A1 = power_of_ten(A1, D1);
	    A2 = atof(buffer+24);
	    D2 = atoi(buffer+24+21);
	    A2 = power_of_ten(A2, D2);
	    A3 = atof(buffer+48);
	    D3 = atoi(buffer+48+21);
	    A3 = power_of_ten(A3, D3);
	    A4 = atof(buffer+72);
	    D4 = atoi(buffer+72+21);
	    A4 = power_of_ten(A4, D4);
	}
    }

    /* "C1" record */
    fread (buffer, 1, 144, fptr);
    int num_categories = atoi(buffer);

    /* "C2" record */
    fread (buffer, 1, 144, fptr);
    char category_name[21];
    strncpy (category_name, buffer, 20);
    category_name[20] = '\0';
    if (coverage())
      coverage()->description(category_name);
    int max_nodes = atoi(buffer+20);
    int actual_nodes = atoi(buffer+26);
    int max_faces = atoi(buffer+32);
    int actual_faces = atoi(buffer+38);
    int max_edges = atoi(buffer+44);
    int actual_edges = atoi(buffer+50);

    /* allocate arrays for primitives */
    MapNode** nodes = new MapNode*[actual_nodes];
    MapFace** faces = new MapFace*[actual_faces];
    MapEdge** edges = new MapEdge*[actual_edges];
    int nnodes = actual_nodes;
    int nfaces = actual_faces;
    int nedges = actual_edges;
    
    int nodenum = 0;
    int facenum = 0;
    int edgenum = 0;
    
    int node_index[max_nodes+1];
    int face_index[max_faces+1];
    int edge_index[max_edges+1];

    int i;
    for (i=0; i<max_nodes; i++) 
	node_index[i] = -1;
    for (i=0; i<max_faces; i++) 
	face_index[i] = -1;
    for (i=0; i<max_edges; i++) 
	edge_index[i] = -1;

    /* set up to construct features */
    boolean area_flag = strcmp(name(), "area") == 0;
    nfeats = area_flag ? actual_faces : actual_edges;
    DlgFeature** features = new DlgFeature*[nfeats];

    /* loop until first line feature, gobbling up area features as we go */
    while (!feof(fptr)) {
	fread (buffer, 1, 144, fptr);
	if (buffer[0] == 'L' && buffer[1] == ' ') break;
	if (buffer[0] != 'A' || buffer[1] != ' ' || !area_flag) continue;

	int id = atoi(buffer+2);
	int nattr = atoi(buffer+20);

	features[facenum] = new DlgFeature((DlgFeatureClass*)this, id, MapFeature::AreaType);

	int nrec = (int)ceil(nattr/12.);
	int major;
	int minor;
	int attrnum = 0;
	for (int rec=0; rec<nrec; rec++) {
	    fread (buffer, 1, 144, fptr);
	    int count = 0;
	    while (count < 12  && attrnum<nattr) {
		major = atoi(buffer+count*12);
		minor = atoi(buffer+count*12+6);
		count++;
		attrnum++;
	    }
	}
	
	features[facenum]->set_attributes(major, minor);
	faces[facenum] = new MapFace(features[facenum]);
	face_index[id] = facenum;
	features[facenum]->face_primitive(faces[facenum]);
	facenum++;
    }
    /* done looping until first line feature, gobbling up area features as we go */

    /* loop until end of file, gobbling up edge primitives */
    boolean first = true;
    while (!feof(fptr)) {
	if (!first) 
	    fread (buffer, 1, 144, fptr);
	else
	    first = false;
	if (buffer[0] != 'L' || buffer[1] != ' ') continue;

	/* interpret edge primitive record */
	int id = atoi(buffer+2);
	int start_node_id = atoi(buffer+8);
	int end_node_id = atoi(buffer+14);
	int left_face_id = atoi(buffer+20);
	int right_face_id = atoi(buffer+26);
	int npts = atoi(buffer+32);
	int nattr = atoi(buffer+38);
	int nchar = atoi(buffer+44);

	/* read in all the edge coordinate pairs */
	float *x = new float[npts];
	float *y = new float[npts];
	int nrec = (int)ceil(npts/12.);
	int ptnum = 0;
	float tmpx, tmpy;
	for (int rec=0; rec<nrec; rec++) {
	    fread (buffer, 1, 144, fptr);
	    int count = 0;
	    while (count < 12  && ptnum<npts) {
		tmpx = (float)atoi(buffer+count*12);
		tmpy = (float)atoi(buffer+count*12+6);
	        x[ptnum] = A1 * tmpx + A2 * tmpy + A3;
		y[ptnum] = A1 * tmpy - A2 * tmpx + A4;
		count++;
		ptnum++;
	    }
	}
	
	nrec = (int)ceil(nattr/12.);
	int major;
	int minor;
	int attrnum = 0;
	for (int rec=0; rec<nrec; rec++) {
	    fread (buffer, 1, 144, fptr);
	    int count = 0;
	    while (count < 12  && attrnum<nattr) {
		major = atoi(buffer+count*12);
		minor = atoi(buffer+count*12+6);
		count++;
		attrnum++;
	    }
	}
	
	MapEdge* edge = new MapEdge;
	edges[edgenum] = edge;
	edges[edgenum]->insert_pointers(npts, x, y, nil, true);
	if (!area_flag) {
	    features[edgenum] = new DlgFeature((DlgFeatureClass*)this, id, MapFeature::LineType);
	    features[edgenum]->set_attributes(major, minor);
	    features[edgenum]->edge_primitive(edges[edgenum]);
	}
	edgenum++;

	MapNode* start_node;
	if (node_index[start_node_id]<0) {
	    node_index[start_node_id] = nodenum;
	    start_node = new MapNode();
	    nodes[nodenum++] = start_node;
	} else 
	    start_node = nodes[node_index[start_node_id]];
	MapNode* end_node;
	if (node_index[end_node_id]<0) {
	    node_index[end_node_id] = nodenum;
	    end_node = new MapNode();
	    nodes[nodenum++] = end_node;
	} else 
	    end_node = nodes[node_index[end_node_id]];
	edge->attach_nodes(start_node, end_node);

	if (area_flag) {
	    MapFace* left_face;
	    if (face_index[left_face_id]<0) {
		face_index[left_face_id] = facenum;
		left_face = new MapFace();
		faces[facenum++] = left_face;
	    } else 
		left_face = faces[face_index[left_face_id]];
	    MapFace* right_face;
	    if (face_index[right_face_id]<0) {
		face_index[right_face_id] = facenum;
		right_face = new MapFace();
		faces[facenum++] = right_face;
	    } else 
		right_face = faces[face_index[right_face_id]];

	    edge->attach_faces(left_face, right_face);
	}
	    
    }
    /* done looping until eof, gobbling up edge primitives */

    do { fread (buffer, 1, 144, fptr); } while (!feof(fptr));
    pclose(fptr);
    return (MapFeature**)features;
}
}

void DlgFeatureClass::extent(double& x1, double& y1, double& x2, double& y2, int& utmzone) {
    if (_extent) {
	x1 = _x1;
	y1 = _y1;
	x2 = _x2;
	y2 = _y2;
	utmzone = _zone;
	return;
    }
    _extent = true;
    if (((DlgDatabase*)(this->coverage()->library()->database()))->Is100K()) {
	FILE* fptr = fopen(path(), "r");
	if (fptr) {
	
	char buffer[83];
	buffer[82] = '\0';
	float swx, swy, nwx, nwy, nex, ney, sex, sey;
	for (int recnum=0; recnum<4; recnum++)
	  fgets (buffer, 82, fptr);
	_zone = atoi(buffer+12);
	do {
	    fgets (buffer, 82, fptr);
	    if (strncmp(buffer, "SW    ", 6)==0)
		break;
	} while (1);
	swx = atof(buffer + 36);
	swy = atof(buffer + 48);
	fgets (buffer, 82, fptr);
	nwx = atof(buffer + 36);
	nwy = atof(buffer + 48);
	fgets (buffer, 82, fptr);
	nex = atof(buffer + 36);
	ney = atof(buffer + 48);
	fgets (buffer, 82, fptr);
	sex = atof(buffer + 36);
	sey = atof(buffer + 48);
	x1 = min(swx, nwx);
	y1 = min(swy, sey);
	x2 = max(nex, sex);
	y2 = max(nwy, ney);
	utmzone = _zone;
	fclose(fptr);
	}
	_x1 = x1;
	_y1 = y1;
	_x2 = x2;
	_y2 = y2;
    }
    else {
    char shcmd[strlen("gunzip -c ") + 1 + strlen(coverage()->path())];
    if (strstr(coverage()->path(), ".Z"))
	strcpy( shcmd, "zcat ");
    else if ( strstr(coverage()->path(), ".gz"))
	strcpy( shcmd, "gunzip -c ");
    else
	strcpy( shcmd, "cat ");
    strcat( shcmd, coverage()->path());

    FILE* fptr  = popen( shcmd, "r");
    if (!fptr)	return;
    char buffer[145];
    buffer[144] = '\0';

    /* "A" records */
    for (int recnum=0; recnum<6; recnum++)
	fread (buffer, 1, 144, fptr);

    /* "B" records */
    double A1, A2, A3, A4;
    int D1, D2, D3, D4;
    int swx, swy, nwx, nwy, nex, ney, sex, sey;
    for (int recnum=0; recnum<2; recnum++) {
	fread (buffer, 1, 144, fptr);
	if (recnum == 0) {
	    A1 = atof(buffer);
	    D1 = atoi(buffer+21);
	    A1 = power_of_ten(A1, D1);
	    A2 = atof(buffer+24);
	    D2 = atoi(buffer+24+21);
	    A2 = power_of_ten(A2, D2);
	    A3 = atof(buffer+48);
	    D3 = atoi(buffer+48+21);
	    A3 = power_of_ten(A3, D3);
	    A4 = atof(buffer+72);
	    D4 = atoi(buffer+72+21);
	    A4 = power_of_ten(A4, D4);
	}
	else if (recnum == 1) { // record B.2
	    swx = atoi(buffer+2);
	    swy = atoi(buffer+2+6);
	    nwx = atoi(buffer+2+6+6+2);
	    nwy = atoi(buffer+2+6+6+2+6);
	    nex = atoi(buffer+2+6+6+2+6+6+2);
	    ney = atoi(buffer+2+6+6+2+6+6+2+6);
	    sex = atoi(buffer+2+6+6+2+6+6+2+6+6+2);
	    sey = atoi(buffer+2+6+6+2+6+6+2+6+6+2+6);
	}
    }
    do { fread (buffer, 1, 144, fptr); } while (!feof(fptr));
    pclose(fptr);

    x1 = _x1 = A1 * swx + A2 * swy + A3;
    y1 = _y1 = A1 * swy - A2 * swx + A4;
    x2 = _x2 = A1 * nex + A2 * ney + A3;
    y2 = _y2 = A1 * ney - A2 * nex + A4;
}
}
