/*
 * Copyright 1994, 1995 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#ifndef dlgcomps_h
#define dlgcomps_h

#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapviews.h>

class DlgCoverage;
class DlgDatabase;
class DlgFeatureClass;
class DlgLibrary;
class MapCoverage;
class MapLibrary;

class DlgFClassComp : public MapFClassComp {
public:
    DlgFClassComp(DlgFeatureClass*);
    DlgFClassComp(istream&);
    virtual ~DlgFClassComp();

    virtual void Interpret(Command*);
    virtual void Uninterpret(Command*);

    virtual Component* Copy();
    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);

    DlgFeatureClass* GetFeatureClass() { return (DlgFeatureClass*)_mapfc; }

protected:
    ParamList* GetParamList();
    void GrowParamList(ParamList*);
    static ParamList* _dlg_fcl_params;
};

class DlgFClassView : public MapFClassView {
public:
    DlgFClassView(DlgFClassComp* = nil);
    virtual ~DlgFClassView();

    virtual void Interpret(Command*);
    virtual void Uninterpret(Command*);

    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);
};

class DlgCovComp : public MapCovComp {
public:
    DlgCovComp(DlgCoverage*);
    DlgCovComp(istream&);
    virtual ~DlgCovComp();

protected:
    virtual MapFClassComp* CreateFClassComp(MapFeatureClass*);

protected:
    ParamList* GetParamList();
    void GrowParamList(ParamList*);
    static ParamList* _dlg_cov_params;
};

class DlgLibComp : public MapLibComp {
public:
    DlgLibComp(DlgLibrary*);
    DlgLibComp(istream&);
    virtual ~DlgLibComp();

protected:
    virtual MapCovComp* CreateCovComp(MapCoverage*);

protected:
    ParamList* GetParamList();
    void GrowParamList(ParamList*);
    static ParamList* _dlg_lib_params;
};

class DlgDbComp : public MapDbComp {
public:
    DlgDbComp(DlgDatabase*);
    DlgDbComp(istream&);
    virtual ~DlgDbComp();

protected:
    virtual MapLibComp* CreateLibComp(MapLibrary*);

protected:
    ParamList* GetParamList();
    void GrowParamList(ParamList*);
    static ParamList* _dlg_db_params;
};

class DlgIdrawComp : public MapIdrawComp {
public:
    DlgIdrawComp();
    DlgIdrawComp(istream&);

    virtual ClassId GetClassId();
    virtual boolean IsA(ClassId);

    Component* Copy();

protected:
    ParamList* GetParamList();
    void GrowParamList(ParamList*);
    static ParamList* _dlg_idraw_params;
};

#endif
