/**************************************************************************
 *   HEADER FILE : LOSFUNC.H - Contains prototypes for the LOS software
 *
 *   AUTHOR      : Kevin Russo, SAIC (Original, 1, 2, 3, 4)
 *
 *   Change     Date          Reason
 *   ------     ----          ------
 *   Original   12 Aug 92     n/a   (file name was los.h)
 *   1          26 Oct 92     overall program update
 *   2          01 Apr 93     renamed file and restructured into three
 *                            separate header files
 *   3          22 Jun 93     modified to conform with XVT 3.0 changes
 *   4          26 Jul 93     removed unused sections
 *
 **************************************************************************/

#ifndef LOS_FUNC_HEADER

#define LOS_FUNC_HEADER
#include "noxvt.h"
/**************************************************************************
 *                     MAIN PROGRAM PROTOTYPES
 **************************************************************************/
#if 0
#ifdef PROTO
long  do_task_win (WINDOW win, EVENT *ep);
long  get_window (int wid, int cid);
void  main (int argc, char *argv[]);
void  invoke (int id, long data);
void  do_E_CHAR (short ch, BOOLEAN shift, BOOLEAN control);
#else
long  do_task_win ();
long  get_window ();
void  main ();
void  invoke ();
void  do_E_CHAR ();
#endif
#endif
/**************************************************************************
 *                  APPLICATION INIT/CLEANUP PROTOTYPES
 **************************************************************************/

#ifdef PROTO
void    do_appl_init (void);
void    do_appl_cleanup (void);
#else
void    do_appl_init ();
void    do_appl_cleanup ();
#endif

/**************************************************************************
 *                        DIALOG PROTOTYPES
 **************************************************************************/
#if 0
#ifdef PROTO
long do_dted_dlg (WINDOW dted_dlg, EVENT *ep);
long do_about_dlg (WINDOW about_dlg, EVENT *ep);
long do_accuracy_dlg (WINDOW accuracy_dlg, EVENT *ep);
long do_pref_dlg (WINDOW pref_dlg, EVENT *ep);
long do_profile_dlg (WINDOW profile_dlg, EVENT *ep);
long do_contrast_dlg (WINDOW contrast_dlg, EVENT *ep);
long do_color_dlg (WINDOW color_dlg, EVENT *ep);
long do_los_dlg (WINDOW los_dlg, EVENT *ep);
#else
long do_dted_dlg ();
long do_about_dlg ();
long do_accuracy_dlg ();
long do_pref_dlg ();
long do_profile_dlg ();
long do_contrast_dlg ();
long do_color_dlg ();
long do_los_dlg ();
#endif

/**************************************************************************
 *                        WINDOW PROTOTYPES
 **************************************************************************/

#ifdef PROTO
long do_basemap_win (WINDOW win, EVENT *ep);
long do_data_win (WINDOW win, EVENT *ep);
long do_profile_win (WINDOW win, EVENT *ep);
long do_status_win (WINDOW win, EVENT *ep);
void invoke_basemap_win (long data);
void invoke_data_win (long data);
void invoke_profile_win (long data);
void invoke_status_win (long data);
#else
long do_basemap_win ();
long do_data_win ();
long do_profile_win ();
long do_status_win ();
void invoke_basemap_win ();
void invoke_data_win ();
void invoke_profile_win ();
void invoke_status_win ();
#endif

/**************************************************************************
 *                        MENU PROTOTYPES
 **************************************************************************/

#ifdef PROTO
void do_file_menu (MENU_TAG tag, BOOLEAN shift, BOOLEAN control);
void do_reset_menu (MENU_TAG tag, BOOLEAN shift, BOOLEAN control);
void do_view_menu (MENU_TAG tag, BOOLEAN shift, BOOLEAN control);
void update_file_menu (BOOLEAN file, BOOLEAN open,  BOOLEAN new,
                       BOOLEAN close, BOOLEAN save, BOOLEAN save_as,
                       BOOLEAN pref, BOOLEAN quit, BOOLEAN about);
void update_view_menu (BOOLEAN view, BOOLEAN settings, BOOLEAN basemap,
                       BOOLEAN profile, BOOLEAN status, BOOLEAN zoom_in,
                       BOOLEAN zoom_out, BOOLEAN contrast, BOOLEAN ch_set,
                       BOOLEAN ch_base, BOOLEAN ch_prof, BOOLEAN ch_stat,
                       BOOLEAN ch_bore);
void update_reset_menu (BOOLEAN reset, BOOLEAN restore,
                       BOOLEAN prev, BOOLEAN clear_prod);
#else
void do_file_menu ();
void do_reset_menu ();
void do_view_menu ();
void update_file_menu ();
void update_view_menu ();
void update_reset_menu ();
#endif
#endif

/**************************************************************************
 *                 LOS INITIALIZATION PROTOTYPES
 **************************************************************************/

#ifdef PROTO
void init_window_information (void);
void init_map_information (void);
void init_mapdefault_settings (void);
void init_system_prefs (void);
void init_percents (void);
void reset_text_edits (void);
void get_default_settings (BOOLEAN all_settings);
void init_handles (void);
void init_cspline_handles (void);
void kill_cspline_handles (void);
void init_trig_tables (void);
#else
void init_window_information ();
void init_map_information ();
void init_map_default_settings ();
void init_system_prefs ();
void init_percents ();
void reset_text_edits ();
void get_default_settings ();
void init_handles ();
void init_cspline_handles ();
void kill_cspline_handles ();
void init_trig_tables ();
#endif

/**************************************************************************
 *                 LOS UTILITY FUNCTION PROTOTYPES
 **************************************************************************/

#ifdef PROTO
ERRSTATUS  load_dted (void);
BOOLEAN    update_settings (BOOLEAN prof, BOOLEAN update);
BOOLEAN    tx_get_setting (short num, char /*far*/ *buf);
BOOLEAN    product_construct (void);
BOOLEAN    product_destruct (FILE *fptr, BOOLEAN destroy);
BOOLEAN    check_dted_file_exists (void);
BOOLEAN    big_endian_machine (void);
BOOLEAN    BoxWithin (EXTENT box1, EXTENT box2);
BOOLEAN    check_file_name (char *filename, char *filetype);
GFLOAT     gt_circle (coord_type pt1, coord_type pt2);
coord_type ChartXY (PNT where);
float      swap_float (float num1);
long       swap_long (long num1);
short      swap_short (short num1);
short      eval (PNT *scrn_pts, short x);
void       dted_destruct (RASTER *raster);
void       sphere_to_xyz (coord_type geo, xyz_type *xyz);
void       flip_radio (WINDOW win, WINDOW ctrl_1_win, WINDOW ctrl_2_win);
void       set_scroll_bars (void);
void       draw_loc_indicators (BOOLEAN obs, BOOLEAN targ);
void       status_update (char *message);
void       fix_settings (void);
void       map_redraw (RASTER *raster, WINDOW map_window);
void       update_range (GFLOAT length);
void       update_elevation (GFLOAT lat, GFLOAT lon);
void       update_vertical_acc (void);
void       makespline (PNT *scrn_pts, long num_pts);
void       DrawMap (char *msg, BOOLEAN load_map);
void       rightjust (char *str);
void       leftjust (char *str);
#else
ERRSTATUS  load_dted();
BOOLEAN    update_settings ();
BOOLEAN    tx_get_setting ();
BOOLEAN    product_construct ();
BOOLEAN    product_destruct ();
BOOLEAN    check_dted_file_exists ();
BOOLEAN    BoxWithin ();
BOOLEAN    check_file_name ();
GFLOAT     gt_circle ();
coord_type ChartXY ();
float      swap_float ();
long       swap_long ();
short      swap_short ();
short      eval();
void       dted_destruct ();
void       sphere_to_xyz ();
void       flip_radio ();
void       set_scroll_bars ();
void       draw_loc_indicators ();
void       status_update ();
void       fix_settings ();
void       map_redraw ();
void       update_range ();
void       update_elevation ();
void       update_vertical_acc ();
void       makespline ();
void       DrawMap ();
void       rightjust ();
void       leftjust ();
#endif

#ifdef PROTO
#if 0 /*XVT_OS == XVT_OS_MAC*/
short double swap_double (short double num1);
#else
double swap_double (double num1);
#endif
#else
#if 0 /*XVT_OS == XVT_OS_MAC*/
short double swap_double ();
#else
double swap_double ();
#endif
#endif

/**************************************************************************
 *                     LOS CALCULATION PROTOTYPES
 **************************************************************************/

#ifdef PROTO
short obsv_elev (GFLOAT lat, GFLOAT lon);
short interp_elev (short pHMEM dted, GFLOAT lat, GFLOAT lon);
int   precise_array ( GFLOAT obsv_lat, GFLOAT obsv_lon,
                      RASTER *elev_raster, GHANDLE *precise_matrix);
void  do_product (void);
void  do_los (void);
void  construct_profile (void);
void  init_ui_params (RASTER **tprod_raster);
void  mem_alloc (long size, GHANDLE *handle, void *pHMEM pointer);
void  make_radials (void);
void  raster_assign (RASTER **tprod_raster);
void  raster2_assign (RASTER *tprod_raster, RASTER **howhigh_raster);
void  vis_mask (short pHMEM elev_hgts, long pHMEM prod_buff_ydim,
                long pHMEM prod_buff_xdim, char *pHMEM profile_color);
void  vip_mask (short pHMEM elev_hgts, long pHMEM prod_buff_ydim,
                long pHMEM prod_buff_xdim, char *pHMEM profile_color);
void  precise_position (short selected_row, short selected_col,
                        RASTER *dted_raster, GFLOAT obsv_lat,
                        GFLOAT obsv_lon, GFLOAT *newlat, GFLOAT *new_lon);
void  display_product (RASTER *product, RASTER **backgnd_raster);
void  restore_basemap (RASTER **backgnd_raster );
#else
short interp_elev ();
short obsv_elev ();
int   precise_array ();
void  do_product ();
void  do_los ();
void  construct_profile ();
void  init_ui_params ();
void  mem_alloc ();
void  make_radials ();
void  raster_assign ();
void  raster2_assign ();
void  vis_mask ();
void  vip_mask ();
void  precise_position ();
void  display_product ();
void  restore_basemap ();
#endif
int set_settings(float startaz, float endaz,
		  float obsx, float obsy, float obsheight, /* float obswidth,*/
		  float tarx, float tary, float tarheight, float tarwidth,
		  float range, long vertacc,
		  int priormask, /* unused? */
		  int prodtype /* enum { VIS, VIP } 2 or 4 val  */,
		  int prof /* viewshed or profile */);

#endif /* LOS_FUNC_HEADER */
