#ifndef H_MUSE
#include "muse.h"
#endif
#include "rgbdefs.h"

#define PACIFY_INC 10

#if XVT_CC_PROTO
MUSE_API raster_construct(FILE * file, RASTER ** pointer)
#else
MUSE_API raster_construct(file, pointer)
    FILE           *file;
    RASTER        **pointer;
#endif
{

    /*
     * Memory allocation and initialization the RASTER structure. If file is
     * not null, the raster is read in.
     */
    RASTER         *raster;
    size_t          items_read = 0;
    unsigned long   length_in_bytes = 0;
    ERRSTATUS       status = STAT_SUCCESS;
    unsigned char   hdr[SZ_RASTER];
    long            temp = SZ_RASTER;
    char            string[80];

#if 0 /* XVTOS == DOSOS*/
    if (temp < sizeof(RASTER))
	printf("Reset SZ_RASTER to:%u\n", sizeof(RASTER));
#endif

    raster = (RASTER *) muse_malloc(sizeof(RASTER), "raster");
    if (raster == NULL)
	return (STAT_MEM_ALLOC_ERR);
    *pointer = raster;
    memset(raster, '\0', sizeof(RASTER));
    raster->magic = MAGIC_RASTER;

    if (file)
    {
	sprintf(string, "read raster: %ld", ftell(file));
	/*muse_log(string);*/
	items_read = fread((void *) hdr, sizeof(hdr), (size_t) 1, file);
	decode_raster_hdr(hdr, raster);

	if (items_read != 1)
	{
	    printf("Unable to read in the raster\n");
	    return (STAT_FILE_READ_ERR);
	}
	if (raster->magic != MAGIC_RASTER)
	{
	    printf("The raster data read in is incorrect.\n");
	    memset(raster, '\0', sizeof(RASTER));
	    return (STAT_DATA_ERR);
	}
	sprintf(string, "raster construct XY %d %d",
		(int) raster->data_matrix.width,
		(int) raster->data_matrix.height);
	/*muse_log(string);*/
	if (raster->data_matrix.ghandle == (GHANDLE) NULL)
	{
	    length_in_bytes = (raster->data_matrix.width
			       * raster->data_matrix.height
			       * raster->data_matrix.bits_per_pixel)
		      / BITS_PER_BYTE;
	    if (!(raster->data_matrix.ghandle = xvt_gmem_alloc(length_in_bytes)))
		return (STAT_MEM_ALLOC_ERR);
	    sprintf(string, "alloc raster->data_matrix size%ld", length_in_bytes);
	    /*muse_log(string);*/
	    status = big_fread(file, raster->data_matrix.ghandle,
			       length_in_bytes,
			       raster->data_matrix.num_type,
			       raster->data_matrix.data_type);
	    if (status)
	    {
                if ( status  == STAT_ABORT )
                {
                     printf("Read aborted by user...\n");
                     return(status);
                }
                else
                {
		     printf("Unable to read in the raster matrix data\n");
		     return (status);
                }
	    }
	}
	if (raster->palette)
	{
	    if ((palette_construct(file, &raster->palette)) != STAT_SUCCESS)
		return STAT_DATA_ERR;
	}
	if (raster->lut)
	{
	    if ((lut_construct(file, raster->palette, &raster->lut)) != STAT_SUCCESS)
		return STAT_DATA_ERR;
	}
    }
    else
    {
	(raster)->palette = (PALETTE *) NULL;
	(raster)->lut = (LUT *) NULL;
	(raster)->data_matrix.width = 2;
	(raster)->data_matrix.height = 1;
	(raster)->data_matrix.ghandle = (GHANDLE) NULL;
    }

    return STAT_SUCCESS;
}				/* end raster_construct */


#if XVT_CC_PROTO
ERRSTATUS
raster_construct_data(RASTER * raster)
#else
ERRSTATUS
raster_construct_data(raster)
    RASTER         *raster;
#endif

/*
 * Memory allocation for the Raster DATA.
 */

{
    unsigned long   length_in_bytes = 0;

    length_in_bytes = raster->data_matrix.width
	      * raster->data_matrix.height
	      * raster->data_matrix.bits_per_pixel
	      / BITS_PER_BYTE;
    if (!(raster->data_matrix.ghandle = xvt_gmem_alloc(length_in_bytes)))
	return (STAT_MEM_ALLOC_ERR);
    return (STAT_SUCCESS);
}

#if XVT_CC_PROTO
MUSE_API raster_destruct(FILE * file, BOOLEAN destruct, RASTER ** raster)
#else
MUSE_API raster_destruct(file, destruct, raster)
    FILE           *file;
    BOOLEAN         destruct;
    RASTER        **raster;
#endif
{
    ERRSTATUS       status = STAT_SUCCESS;
    unsigned long   length_in_bytes = 0;
    size_t          items_written = 0;
    unsigned char   hdr[SZ_RASTER];
    char            string[80];

    if (file)
    {
	encode_raster_hdr(hdr, *raster);
	sprintf(string, "writing raster: %ld", ftell(file));
	/* muse_log(string); */
	items_written = fwrite((void *) hdr, sizeof(hdr), (size_t) 1, file);
	if (items_written != 1)
	{
	    printf("Unable to write the raster\n");
	    return (STAT_FILE_WRITE_ERR);
	}
        
	if ((*raster)->data_matrix.ghandle != (GHANDLE) NULL)
	{
	    length_in_bytes = ((*raster)->data_matrix.width
			       * (*raster)->data_matrix.height
			       * (*raster)->data_matrix.bits_per_pixel)
		               / BITS_PER_BYTE;
	    status = big_fwrite(file,
				(*raster)->data_matrix.ghandle,
				(unsigned long) length_in_bytes,
				(*raster)->data_matrix.num_type,
				(*raster)->data_matrix.data_type);

	    if (status)
	    {
		printf("Unable to write the raster matrix data\n");
		return (STAT_FILE_WRITE_ERR);
	    }
	}
    }

    if ((*raster)->palette != (PALETTE *) NULL && (*raster)->palette != (PALETTE *) SZ_PALETTE)
	palette_destruct(file, destruct, &(*raster)->palette);
    if ((*raster)->lut != (LUT *) NULL && (*raster)->lut != (LUT *) SZ_LUT)
	lut_destruct(file, destruct, &(*raster)->lut);

    
    if (destruct)
    {
	if ((*raster)->data_matrix.ghandle)
	{
	    if (!xvt_gmem_free((*raster)->data_matrix.ghandle))
	    {
		printf("Error in raster_destruct\n");
		(*raster)->data_matrix.ghandle = (GHANDLE) NULL;
	    }
	    sprintf(string, "free raster->data_matrix");
	    /* muse_log(string); */

	}
        
	muse_free(*raster, "raster");
	*raster = (RASTER *) NULL;
    }
    return (status);
}				/* end raster_destruct */

#if XVT_CC_PROTO
ERRSTATUS
raster_setup(BASEMAP * basemap, MGM * mgm)
#else
ERRSTATUS
raster_setup(basemap, mgm)
    BASEMAP        *basemap;
    MGM            *mgm;
#endif
{
    /* set the basemap params */
    basemap->product = BASE_OTHER;
    strcpy((char *) &basemap->file_spec.type, FT_RAS);
    strcpy((char *) &basemap->file_spec.name, "*.ras");
    /*    xvt_fsys_get_dir(&basemap->file_spec.dir);*/
    basemap->render_info.shade_threshold = 25.0;
    basemap->render_info.precision = 10.0;
    basemap->render_info.sunx = -1;
    basemap->render_info.suny = 1;
    basemap->needs_reload = TRUE;
    NOREF(mgm);
    return STAT_SUCCESS;
}


#if XVT_CC_PROTO
MUSE_API raster_xform(
	     RASTER * in,
	     RASTER * out,
	     RENDER_INFO * render_info,
	     LUT * default_lut,
	     PALETTE * default_palette)
#else
MUSE_API raster_xform(in, out, render_info, default_lut, default_palette)
    RASTER         *in;
    RASTER         *out;
    RENDER_INFO    *render_info;
    LUT            *default_lut;
    PALETTE        *default_palette;
#endif
{
    ERRSTATUS       status = STAT_SUCCESS;

    switch (in->data_matrix.data_type)
    {
      case DATA:
	switch (out->data_matrix.data_type)
	{
	  case DATA:

	    /*
	     * data to data functions might at the most, possibly rescale the
	     * data or convert the numeric type, from say integer to float.
	     */
	    status = raster_copy(in, out);
	    break;

	  case COLOR_RGB:

	    /*
	     * render the data into an RGB image.  This is where we will put
	     * the continuous tone renderer
	     */
	    printf("Unsupported case in raster_xform.\n");

	    break;

	  case COLOR_INDIRECT:

	    /*
	     * render the data into an indirect image. Used to display DTED
	     */
	    status = raster_data_to_indirect(in, out, render_info, default_lut, default_palette);
	    break;

	  default:
	    printf("Unsupported case in raster_xform.\n");
	    break;
	}
	break;

      case COLOR_RGB:
	switch (out->data_matrix.data_type)
	{
	  case DATA:

	    /*
	     * this would convert an RGB into a data value. A possible
	     * application would be image classification
	     */
	    break;

	  case COLOR_RGB:

	    /*
	     * General image processing functions such as histogram
	     * equalization go here.
	     */
	    status = raster_copy(in, out);
	    break;

	  case COLOR_INDIRECT:

	    /*
	     * apply minimim distance color classification Used to display
	     * ADRG on non-RGB systems.
	     */
	    status = raster_rgb_to_indirect(in, out, default_palette);
	    break;

	  default:
	    printf("Unsupported case in raster_xform.\n");
	    break;
	}
	break;

      case COLOR_INDIRECT:
	switch (out->data_matrix.data_type)
	{
	  case DATA:
	    /* this case would xform COLOR_INDIRECT into DATA */
	    break;

	  case COLOR_RGB:

	    /*
	     * this case would index color numbers through a palette to get
	     * RGB values
	     */
	    status = raster_indirect_to_rgb(in, out);
	    if (status)
		printf("Unable to convert image to RGB\n");
	    break;

	  case COLOR_INDIRECT:

	    /*
	     * This case would to at most change the number of colors in the
	     * image.
	     */

	    status = raster_indirect_to_indirect(in, out, default_palette);
	    if (status)
		printf("Unable to convert the indirect image\n");
	    break;

	  default:
	    printf("Unsupported case in raster_xform.\n");
	    break;
	}			/* end switch out->data_matrix.data_type */
	break;

      default:
	printf("Unsupported case in raster_xform.\n");
	break;
    }				/* end switch */

    return (status);
}

#if XVT_CC_PROTO
ERRSTATUS
raster_copy(RASTER * in, RASTER * out)
#else
ERRSTATUS
raster_copy(in, out)
    RASTER         *in;
    RASTER         *out;
#endif
{
    ERRSTATUS       status = STAT_SUCCESS;

    BYTE pHMEM      in_data;
    BYTE pHMEM      out_data;
    long            in_len, out_len, ilong;


    if (in->data_matrix.data_type != out->data_matrix.data_type)
    {
	printf("Data types muse match for raster_copy.\n");
	return (-1);
    }
    if (in->data_matrix.num_type != out->data_matrix.num_type)
    {
	printf("Number types muse match for raster_copy.\n");
	return (-1);
    }
#if 0
    in_len = xvt_gmem_get_size(in->data_matrix.ghandle);
    out_len = xvt_gmem_get_size(out->data_matrix.ghandle);
    if (in_len > out_len)
	printf("in_data > out_data\n");
#endif
    /* lock the data for access */
    in_data = (BYTE pHMEM) xvt_gmem_lock(in->data_matrix.ghandle);
    if (in_data == (BYTE pHMEM) NULL)
	printf("in_data handle is null\n");

    out_data = (BYTE pHMEM) xvt_gmem_lock(out->data_matrix.ghandle);
    if (out_data == (BYTE pHMEM) NULL)
	printf("out_data handle is null\n");
#if 0
    if (xvt_gmem_get_size(in->data_matrix.ghandle) > xvt_gmem_get_size(out->data_matrix.ghandle))
	printf("in_data > out_data\n");
    else
#endif
	for (ilong = 0; ilong < in_len; ilong++)
	    out_data[ilong] = in_data[ilong];
    NOREF(status);
    return (STAT_SUCCESS);
}


#if XVT_CC_PROTO
ERRSTATUS
raster_indirect_to_indirect(RASTER * in, RASTER * out,
			    PALETTE * default_palette)
#else
ERRSTATUS
raster_indirect_to_indirect(in, out, default_palette)
    RASTER         *in;
    RASTER         *out;
    PALETTE        *default_palette;
#endif
{
    ERRSTATUS       status = STAT_SUCCESS;

    BYTE pHMEM      in_data;
    BYTE pHMEM      out_data;
    long            in_len, out_len, index;
    USHORT          map[MAX_COLORS];


    if (in->data_matrix.data_type != out->data_matrix.data_type)
    {
	printf("Data types muse match for indirect_to_indirect.\n");
	return (-1);
    }
    if (in->data_matrix.num_type != out->data_matrix.num_type)
    {
	printf("Number types muse match for indirect_to_indirect.\n");
	return (-1);
    }
#if 0
    in_len = xvt_gmem_get_size(in->data_matrix.ghandle);
    out_len = xvt_gmem_get_size(out->data_matrix.ghandle);
    if (in_len > out_len)
	printf("in_data > out_data\n");
#endif
    if (out->palette)
       palette_destruct (NULL, TRUE, &(out->palette));
    status = palette_construct((FILE *) NULL, &out->palette);
    if (status)
	printf("Out of memory");
    memcpy(out->palette, default_palette, sizeof(PALETTE));
    for (index = 0; index < in->palette->num_colors; index++)
    {
	status = pal_nearest_color(out->palette,
				   in->palette->rgb[index],
				   &map[index]);
	if (status)
	    printf("Pal_nearest_color failed\n");
    }
    /* lock the data for access */
    in_data = (BYTE pHMEM) xvt_gmem_lock(in->data_matrix.ghandle);
    if (in_data == (BYTE pHMEM) NULL)
	printf("in_data handle is null\n");

    out_data = (BYTE pHMEM) xvt_gmem_lock(out->data_matrix.ghandle);
    if (out_data == (BYTE pHMEM) NULL)
	printf("out_data handle is null\n");
#if 0
    if (xvt_gmem_get_size(in->data_matrix.ghandle) > xvt_gmem_get_size(out->data_matrix.ghandle))
	printf("in_data > out_data\n");
    else
#endif
	for (index = 0; index < in_len; index++)
	    out_data[index] = map[in_data[index]] + PAL_OFFSET;

    return (STAT_SUCCESS);
}

#if XVT_CC_PROTO
ERRSTATUS
raster_rgb_to_indirect(
		       RASTER * in,
		       RASTER * out,
		       PALETTE * default_palette)
#else
ERRSTATUS
raster_rgb_to_indirect(in, out, default_palette)
    RASTER         *in;
    RASTER         *out;
    PALETTE        *default_palette;
#endif
{
    ERRSTATUS       status = STAT_SUCCESS;
    BYTE pHMEM      in_data;
    BYTE pHMEM      out_data;
    BYTE pHMEM      temp;
    int             bits;
    unsigned long   data_index = 0;
    int             colr[8];
    short           x, y;
    USHORT          color_index;
    long            i = 0;
    long            pix_per_byte;
/*    long            c;*/
    RGB             rgb;

    /* check for proper data forms */
    if (in->data_matrix.data_type != COLOR_RGB)
    {
	printf("The input raster should have contained COLOR_RGB.\n");
	return (-1);
    }
    if (out->data_matrix.data_type != COLOR_INDIRECT)
    {
	printf("The output raster should have been set to receive COLOR_INDIRECT.\n");
	return (-1);
    }
    /* input RGB raster must have pal */
    if (in->palette == (PALETTE *) NULL)
    {
	in->palette = (PALETTE *) muse_malloc(sizeof(PALETTE), "RASTER.C in->palette\n");
	if (in->palette == (PALETTE *) NULL)
	    printf("Out of memory");
	memcpy(in->palette, default_palette, sizeof(PALETTE));
    }

    if (out->palette)
       palette_destruct (NULL, TRUE, &(out->palette));
    out->palette = (PALETTE *) muse_malloc(sizeof(PALETTE), "RASTER.C out->palette\n");
    if (out->palette == (PALETTE *) NULL)
        printf("out of memory");
    memcpy(out->palette, in->palette, sizeof(PALETTE));

    /* lock the data for access */
    in_data = (BYTE pHMEM) xvt_gmem_lock(in->data_matrix.ghandle);
    if (in_data == (BYTE pHMEM) NULL)
	printf("in_data handle is null\n");
    out_data = (BYTE pHMEM) xvt_gmem_lock(out->data_matrix.ghandle);
    if (out_data == (BYTE pHMEM) NULL)
	printf("out_data handle is null\n");

    pix_per_byte = BITS_PER_BYTE / out->data_matrix.bits_per_pixel;

    /* render the surface, pixel by pixel into the output raster */
/*    c = 0;*/
    temp = in_data;

    for (y = 0; y < in->data_matrix.height; y++)
    {
	for (x = 0; x < in->data_matrix.width; x += pix_per_byte, i++)
	{
	    for (bits = 0; bits < pix_per_byte; data_index++, bits++)
	    {
		rgb.r = *temp++ << 8;
		rgb.g = *temp++ << 8;
		rgb.b = *temp++ << 8;
		status = pal_nearest_color(out->palette, rgb, &color_index);
		if (status)
		{
		    printf("Error getting nearest palette color\n");
		    return (-1);
		}
		colr[bits] = color_index;
	    }
	    switch (pix_per_byte)
	    {
	      case 1:
		out_data[i] = (BYTE) (colr[0]) + PAL_OFFSET;

		break;
	      case 8:
		out_data[i] = (BYTE)
			  ((colr[0] << 7) & 0x80) |
			  ((colr[1] << 6) & 0x40) |
			  ((colr[2] << 5) & 0x20) |
			  ((colr[3] << 4) & 0x10) |
			  ((colr[4] << 3) & 0x08) |
			  ((colr[5] << 2) & 0x04) |
			  ((colr[6] << 1) & 0x02) |
			  ((colr[7] << 0) & 0x01);
		break;
	      case 2:
		out_data[i] = (BYTE)
			  ((colr[0] << 4) & 0xf0) |
			  (colr[1] & 0x0f);
		break;
	    }			/* end of switch */
	}			/* end for x */
/*	if (y % PACIFY_INC == 0)
	{
	    if (pacify("Computing the image.", (short)
		       ((float) y / (float) in->data_matrix.height * 100.0)))
	    {
		pacify("Ready", (short) NULL);
		return (STAT_ABORT);
	    }
	}*/
    }				/* end for y */
/*    pacify("Ready", (short) NULL);*/

    xvt_gmem_unlock(in->data_matrix.ghandle);
    xvt_gmem_unlock(out->data_matrix.ghandle);
    return (status);
}

#if XVT_CC_PROTO
ERRSTATUS
raster_data_to_indirect(
			RASTER * in,
			RASTER * out,
			RENDER_INFO * render_info,
			LUT * default_lut,
			PALETTE * default_palette)
#else
ERRSTATUS
raster_data_to_indirect(in, out,
			render_info, default_lut, default_palette)
    RASTER         *in;
    RASTER         *out;
    RENDER_INFO    *render_info;
    LUT            *default_lut;
    PALETTE        *default_palette;
#endif
{
    short pHMEM     in_data;
    BYTE pHMEM      out_data;
    GFLOAT          delta;
    int             bits;
    unsigned long   data_index = 0;
    int             colr[8];
    USHORT          x, y, color_index;
    long            i = 0;
    long            pix_per_byte;
    /* check for proper data forms */
    if (in->data_matrix.data_type != DATA)
    {
	printf("The input raster should have contained DATA.\n");
	return (-1);
    }
    if (out->data_matrix.data_type != COLOR_INDIRECT)
    {
	printf("The output raster should have been set to receive COLOR_INDIRECT.\n");
	return (-1);
    }
    /* muse have lut */
    if (in->lut == (LUT *) NULL)
    {
	    in->lut = (LUT *) muse_malloc(sizeof(LUT), "RASTER.C in->lut");
	    if (in->lut == (LUT *) NULL)
	        printf("Out of memory");
	    memcpy(in->lut, default_lut, sizeof(LUT));
        if (in->lut->stretch)
        {
	        /*
	         * Adjust the lut data ranges to the input data. This assumes
	         * that the default lut has 2 ranges, 1 for sea, 1 for land
	         */
	        if (in->data_hdr.data_min < 0)
	    	in->lut->lut_desc[1].data_range[0] =
	    		  in->data_hdr.data_min;
	        else
	    	in->lut->lut_desc[0].data_range[0] =
	    		  in->data_hdr.data_min;
    
	        if (in->data_hdr.data_max < 0)
	    	in->lut->lut_desc[1].data_range[1] =
	    		  in->data_hdr.data_max;
	        else
	    	in->lut->lut_desc[0].data_range[1] =
	    		  in->data_hdr.data_max;
        }
    }				/* end if */
    /* hang the default palette onto the output image */
    if (out->palette)
       palette_destruct (NULL, TRUE, &(out->palette));
    out->palette = (PALETTE *) muse_malloc(sizeof(PALETTE), "RASTER.C out->palette");
    if (out->palette == (PALETTE *) NULL)
	printf("Out of memory\n");
    memcpy(out->palette, default_palette, sizeof(PALETTE));

    /* lock the data for access */
    in_data = (short pHMEM) xvt_gmem_lock(in->data_matrix.ghandle);
    if (in_data == (short pHMEM) NULL)
	printf("in_data handle is null\n");
    out_data = (BYTE pHMEM) xvt_gmem_lock(out->data_matrix.ghandle);
    if (out_data == (BYTE pHMEM) NULL)
	printf("out_data handle is null\n");

    pix_per_byte = BITS_PER_BYTE / out->data_matrix.bits_per_pixel;


    /* render the surface, pixel by pixel into the output raster */
    for (y = 0; y < in->data_matrix.height; y++)
    {
	for (x = 0; x < in->data_matrix.width; x += pix_per_byte, i++)
	{
	    for (bits = 0; bits < pix_per_byte; data_index++, bits++)
	    {
		if (x > 0 &&
		    x < (in->data_matrix.width - 1) &&
		    y > 0 &&
		    y < (in->data_matrix.height - 1))
		    delta = in_data[(y + render_info->suny) *
				    in->data_matrix.width +
				    (x + render_info->sunx)]
			      - in_data[data_index];
		else
		    delta = 0;

		lut_lookup(in->lut, (GFLOAT) in_data[data_index],
			   delta, render_info->shade_threshold,
			   render_info->precision, &color_index);
		colr[bits] = (int) color_index;
	    }
	    switch (pix_per_byte)
	    {
	      case 1:
		out_data[i] = (BYTE) (colr[0]) + PAL_OFFSET;
		break;
	      case 8:
		out_data[i] = (BYTE)
			  ((colr[0] << 7) & 0x80) |
			  ((colr[1] << 6) & 0x40) |
			  ((colr[2] << 5) & 0x20) |
			  ((colr[3] << 4) & 0x10) |
			  ((colr[4] << 3) & 0x08) |
			  ((colr[5] << 2) & 0x04) |
			  ((colr[6] << 1) & 0x02) |
			  ((colr[7] << 0) & 0x01);
		break;
	      case 2:
		out_data[i] = (BYTE)
			  ((colr[0] << 4) & 0xf0) |
			  (colr[1] & 0x0f);
		break;
	    }			/* end of switch */
	}			/* end for x */
/*	if (y % PACIFY_INC == 0)
	{
	    if (pacify("Computing the image.", (short)
		       ((float) y / (float) in->data_matrix.height * 100.0)))
	    {
		pacify("Ready", (short) NULL);
		return (STAT_ABORT);
	    }
	}*/
    }				/* end for y */
/*    pacify("Ready", (short) NULL);*/

    xvt_gmem_unlock(in->data_matrix.ghandle);
    xvt_gmem_unlock(out->data_matrix.ghandle);
    return (STAT_SUCCESS);
}				/* end function: raster_data_to_indirect */

#if XVT_CC_PROTO
ERRSTATUS
raster_indirect_to_rgb(
		       RASTER * in,
		       RASTER * out)
#else
ERRSTATUS
raster_indirect_to_rgb(in, out)
    RASTER         *in;
    RASTER         *out;
#endif
{
    ERRSTATUS       status = STAT_SUCCESS;
    BYTE pHMEM      in_data;
    RGB24 pHMEM     out_data;
    unsigned long   data_index = 0;
    short           x, y;


    /* check for proper data forms */
    if (in->data_matrix.data_type != COLOR_INDIRECT)
    {
	printf("The input raster should have contained COLOR_INDIRECT.\n");
	return (-1);
    }
    if (out->data_matrix.data_type != COLOR_RGB)
    {
	printf("The output raster should have been set to receive COLOR_RGB.\n");
	return (-1);
    }
    /* input indirect raster must have pal */
    if (in->palette == (PALETTE *) NULL)
    {
	printf("The input raster should have had a palette.\n");
	return (-1);
    }
    /* lock the data for access */
    in_data = (BYTE pHMEM) xvt_gmem_lock(in->data_matrix.ghandle);
    if (in_data == (BYTE pHMEM) NULL)
	printf("in_data handle is null\n");
    out_data = (RGB24 pHMEM) xvt_gmem_lock(out->data_matrix.ghandle);
    if (out_data == (RGB24 pHMEM) NULL)
	printf("out_data handle is null\n");




    /* image conversion, pixel by pixel into the output raster */
    for (y = 0; y < in->data_matrix.height; y++)
    {
	for (x = 0; x < in->data_matrix.width; x++)
	{
	    out_data[data_index].r = in->palette->rgb[in_data[data_index]].r / 257;
	    out_data[data_index].g = in->palette->rgb[in_data[data_index]].g / 257;
	    out_data[data_index].b = in->palette->rgb[in_data[data_index]].b / 257;
	    data_index++;
	}			/* end for x */
	if (y % PACIFY_INC == 0)
	{
	    if (pacify("Converting the image.", (short)
		       ((float) y / (float) in->data_matrix.height * 100.0)))
	    {
		pacify("Ready", (short) NULL);
		return (STAT_ABORT);
	    }
	}
    }				/* end for y */
    pacify("Ready", -1);

    xvt_gmem_unlock(in->data_matrix.ghandle);
    xvt_gmem_unlock(out->data_matrix.ghandle);
    return (status);
}				/* end function raster_indirect_to_rgb */


/* use NAM space */

#if XVT_CC_PROTO
ERRSTATUS
rgb2uvw(long r, long g, long b, long *u, long *v, long *w)
#else
ERRSTATUS
rgb2uvw(r, g, b, u, v, w)
    long            r, g, b;
    long           *u, *v, *w;
#endif
{
    *u = (long) (1190 * r - 910 * g - 280 * b + (0x00000001 << 5)) >> 6;
    *v = (long) (724 * r + 242 * g - 966 * b + (0x00000001 << 5)) >> 6;
    *w = (long) (1633 * r + 4073 * g + 2470 * b + (0x00000001 << 5)) >> 6;
    return STAT_SUCCESS;
}

#if XVT_CC_PROTO
ERRSTATUS
uvw2rgb(long u, long v, long w, long *r, long *g, long *b)
#else
ERRSTATUS
uvw2rgb(u, v, w, r, g, b)
    long            u, v, w;
    long           *r, *g, *b;
#endif
{
    long            rs, gs, bs;

    rs = (9823 * u + 2400 * v + 2052 * w + (0x00000001 << 17)) >> 18;
    gs = (-7294 * u + 7361 * v + 2052 * w + (0x00000001 << 17)) >> 18;
    bs = (5535 * u - 13725 * v + 2052 * w + (0x00000001 << 17)) >> 18;
    *r = (rs < 0) ? 0 : ((rs > 255) ? 255 : rs);
    *g = (gs < 0) ? 0 : ((gs > 255) ? 255 : gs);
    *b = (bs < 0) ? 0 : ((bs > 255) ? 255 : bs);
    return STAT_SUCCESS;
}

